package edu.toronto.cs.expr;

import static edu.toronto.cs.expr.BoolOp.*;

/** 
 * Choice operator: {@literal ch (a, b)}.
 * <p>
 * {@literal ch (a, b)} could be interpreted as "{@literal true} if {@literal
 * a} is true, {@literal false} if {@literal b} is true".
 * 
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class ChoiceOp extends NamedOp
{
  public static final ChoiceOp CH = new ChoiceOp ("ch", 2);

  private ChoiceOp (String name, int arity)
  {
    super (name, arity);
  }

  public static final boolean isChoice (Expr e)
  {
    return e.op () == CH;
  }

  public static final Expr choice (Expr e1, Expr e2)
  {
    return e1.getFactory ().op (CH).binApply (e1, e2);
  }

  public static final Expr lchoice (Expr e1, Expr e2)
  {
    // ch (t, f) = t
    if (isTrue (e1) && isFalse (e2))
      return e1.getFactory ().op (TRUE);
    // ch (f, t) = f
    if (isFalse (e1) && isTrue (e2))
      return e1.getFactory ().op (FALSE);
    // ch (a, !a) = a
    if (isNot (e2) && e1 == e2.arg (0) || isNot (e1) && e2 == e1.arg (0))
      return e1;

    return choice (e1, e2);
  }

  public static final Expr lgetT (Expr e)
  {
    if (isChoice (e))
      return e.arg (0);

    return e;
  }

  public static final Expr lgetF (Expr e)
  {
    if (isChoice (e))
      return e.arg (1);

    return lnot (e);
  }
}

