package edu.toronto.cs.expr;

/** 
 * Defines a function application Expr.
 *
 * ((Expr function) (Expr arg))
 * 
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class FunctionApplOp extends NamedOp
{
  public static final FunctionApplOp APPL =
    new FunctionApplOp ("function-call", 2);

  public FunctionApplOp (String name, int arity)
  {
    super (name, arity);
  }

  /** 
   * Returns the function of a function application.
   * 
   * @param e 
   * @return 
   */
  public static Expr function (Expr e)
  {
    return e.arg (0);
  }

  /** 
   * Returns the argument of a function application. 
   * 
   * @param e 
   * @return 
   */
  public static Expr arg (Expr e)
  {
    return e.arg (1);
  }

  /** 
   * Returns true iff the given expression is a function application. 
   * 
   * @param e 
   * @return 
   */
  public static boolean isAppl (Expr e)
  {
    return e.op () == APPL;
  }

  /**
   * Returns a function call {@code foo( Expr )}
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr apply (Expr e1, Expr e2)
  {
    return e1.getFactory ().op (APPL).binApply (e1, e2);
  }
}

