package edu.toronto.cs.expr;

/**
 * An interface representing an operator of an expression.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public interface Operator
{
  /**
   * The arity of this operator.
   *
   * @return the arity of this operator
   */
  int arity ();
  
  /**
   * Checkes whether an array of arguments conforms to well-formedness
   * rules of this operator. This is a shallow check that assumes that
   * the arguments are well-formed.
   *
   * @param args an array of arguments
   * @return true if this operator is well-formed with these arguments
   * provided that the arguments are well formed themselves
   */
  boolean isWellFormed (Expr[] args);
  
  /**
   * The syntactic name of this operator.
   *
   * @return the name of this operator as a String
   */
  String name ();
  
  
}
