package edu.toronto.cs.expr;

import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;

/** 
 * A rewriter that simplifies expressions by applying various equivalence
 * rules, including those of Boolean logic, arithmetic, and C.
 * 
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class SimplifyingRewriter extends IdentityRewriter
{
  public Expr rewrite (Expr e)
  {
    e = super.rewrite (e);

    if (isNot (e))
      return lnot (e.arg (0));
    else if (isAnd (e))
      return land (e.arg (0), e.arg (1));
    else if (isOr (e))
      return lor (e.arg (0), e.arg (1));
    else if (e.op () == EQ)
      return lequal (e.arg (0), e.arg (1));
    else if (isGt (e))
      return lgt (e.arg (0), e.arg (1));
    else if (isLt (e))
      return llt (e.arg (0), e.arg (1));
    else if (isDeref (e) && isAddrOf (e.arg (0)))
      return (e.arg (0).arg (0));
    else if (isAddrOf (e) && isDeref (e.arg (0)))
      return (e.arg (0).arg (0));
    else
      return e;
  }
}


