package edu.toronto.cs.expr;

import junit.framework.*;

import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.NumericOp.*;

public class SimplifyingRewriterTestCase extends TestCase
{
  private Expr a, x, y, trueEx, falseEx, e;
  private ExprFactory fac;
  private SimplifyingRewriter simp;

  protected void setUp ()
  {
    fac = new ExprFactoryImpl ();
    simp = new SimplifyingRewriter ();
    trueEx = fac.op (TRUE);
    falseEx = fac.op (FALSE);
    a = fac.var ("a");
    x = fac.var ("x");
    y = fac.var ("y");
    e = null;
  }

  protected void tearDown ()
  {
  }

  public void testAnd ()
  {
    // (TRUE AND a) == a
    e = and (trueEx, a);
    assertEquals (a, simp.rewrite (e));
  }

  public void testOr ()
  {
    // (FALSE OR a) == a
    e = or (falseEx, a);
    assertEquals (a, simp.rewrite (e));
  }
    
  public void testBoolOps ()
  {
    // !!(TRUE AND (FALSE OR a)) = a
    e = not (not (and (trueEx, or (falseEx, a))));
    assertEquals (a, simp.rewrite (e));
  }

  public void testMixedOps ()
  {
    // ((x + y) > y) AND (FALSE OR a) == (x > y) AND a
    Expr xpy_gt_y = gt (plus (x, y), y);
    e = and (xpy_gt_y, or (falseEx, a));
    assertEquals (and (xpy_gt_y, a), simp.rewrite (e));
  }
}

