package edu.toronto.cs.expr;

/** 
 * An interface for guarded substitution. A guarded substitution is a
 * generalization of substitution that decides whether an expression
 * permits substitution. See {@link edu.toronto.cs.expr.Expr#guardedSubst}
 * 
 * @author <a href="mailto:kelvin@cs.toronto.edu">Kelvin Ku</a>
 * @version 1.0
 */
public interface SubstGuard
{
  /** 
   * Returns true iff it's ok to apply variable substitution to the given
   * expression. 
   * 
   * @param e the expression to test for substitutability.
   * @return true if substitution must be applied to sub-expressions
   * of {@code e}, false if it should not.
   */
  boolean okSubst (Expr e);
}
