/**
   Provides classes and interface to represent and manipulate
   syntactic expressions. A syntactic expression is effectively an
   Abstract Syntax Tree (AST) whose nodes are called expressions. 

   <p>
   An expression can have zero or more children, which are called
   <i>arguments</i>. An <i>arity</i> of an expression is the number of
   arguments it can have.
   </p>

   <p>
   An <i>operator</i> is the sole attribute of an expression. An operator
   indicates the semantic meaning of the expression, and posses
   constraints on well-formedness and well-typeness. The only two
   required attributes of an operator are {@code name} that determines
   its String representation, and {@code arity} that determines the
   number of children an expression containing this operator must
   have. 
   </p>

   <p> The following figure shows an example expression tree:<br><br>
   <img src="doc-files/expr_example.png">

   <p> This package provides three interfaces: {@link
   edu.toronto.cs.expr.Expr} to represent expressions, {@link
   edu.toronto.cs.expr.ExprFactory} to represent a factory for
   constructing expressions, and {@link edu.toronto.cs.expr.Operator}
   to represent an operator. See documentation for these interfaces for
   more information.  </p>

   <p> The package also provides an implementation of these interfaces
   in the form of {@link edu.toronto.cs.expr.ExprFactoryImpl} and
   {@link edu.toronto.cs.expr.ExprFactoryImpl.ExprImpl}. A large
   collection of useful operators is provided as well.  </p>

   <p> The operators are arranged in a class hierarchy to reflect
   their semantic meaning. For example, all boolean logical operators,
   such as conjunction and disjunction, are defined in {@link
   edu.toronto.cs.expr.BoolOp}. All operators that define types are
   subclasses of {@link edu.toronto.cs.expr.TypeOp}, etc.  </p>

   <p>
   By convention, a class defining an operator must provide static
   methods: (a) to check if an operator is a top-level operator of an
   expression, (b) to the attributes of the operator and
   sub-expressions, and (c) to construct an expressions from
   sub-expressions. See {@code BoolOp} for an example.
   </p>

   <p>
   All expressions constructed by an expression factory are assumed to
   be canonical. Thus, if {@code e1} and {@code e2} are two
   expressions constructed by a factory, then {@code e1.equals (e2)}
   if and only if {@code e1 == e2}. This requirement may be relaxed in
   the future to allow for canonical and non-canonical factory
   implementations.
   </p>

   <p>
   In addition, the package provides classes to convert between
   expressions and their {@code libcvcl} representation. These classes
   do not fit with the logical intention of this package and will be
   moved elsewhere in the future.
   </p>


 */
package edu.toronto.cs.expr;
