package edu.toronto.cs.yasm;

import edu.toronto.cs.util.StopWatch;
import java.util.*;

/**
 * YasmStatistics.java
 *
 *
 * Created: Wed Jul  7 12:32:03 2004
 *
 * @author <a href="mailto:maxin@epoch.cs">Xin Ma</a>
 * @version 1.0
 */
public class YasmStatistics 
{
  /**
   * measures overall time
   *
   */
  StopWatch overall;
  
  /**
   * time took to compute H function
   *
   */
  StopWatch refine;

  /**
   *predicateAbstraction time
   *
   */
  StopWatch predicateAbstraction;
  /**
   * model-checking time
   *
   */
  StopWatch mc;

  /**
   * Time it took to find new predicates;
   *
   */
  StopWatch newPred;

  /**
   * Time taken to compute symbolic representation of the model
   *
   */
  StopWatch compile;

  /**
   * Parsing time
   *
   */
  StopWatch parse;

  /**
   * WP computation time
   */
  StopWatch wp;

  /**
   * Theorem prover query-only time
   */
  StopWatch tpQuery;

  StopWatch tpAssert;

  StopWatch tpIndices;

  StopWatch tpWriteQuery;

  /**
   * CToAddrModel.rewrite () time
   */
  StopWatch addrModel;

  StopWatch relatedPredicates;

  int numQueries;

  public YasmStatistics () 
  {
    overall = new StopWatch ();
//    queryTimes = new LinkedList<String> ();
  } // YasmStatistics constructor
  
  public void start ()
  {
    overall.reset ();
  }
  public void stop ()
  {
    overall.stop ();
  }

  public void startPredicateAbstraction ()
  {
    if (predicateAbstraction == null)
      predicateAbstraction = new StopWatch ();
    else
      predicateAbstraction.resume ();
  }
  public void stopPredicateAbstraction ()
  {
    predicateAbstraction.pause ();
  }

  public void startCompile ()
  {
    if (compile == null)
      compile = new StopWatch ();
    else
      compile.resume ();
  }
  public void stopCompile ()
  {
    compile.pause ();
  }
  

  public void startParse ()
  {
    if (parse == null)
      parse = new StopWatch ();
    else
      parse.resume ();
  }
  public void stopParse ()
  {
    parse.pause ();
  }
  


  public void startRefine ()
  {
    if (refine == null)
      refine = new StopWatch ();
    else
      refine.resume ();
  }
  public void stopRefine ()
  {
    refine.pause ();
  }

  public void startMC ()
  {
    if (mc == null)
      mc = new StopWatch ();
    else
      mc.resume ();
  }
  public void stopMC ()
  {
    mc.pause ();
  }

  public void startNewPred ()
  {
    if (newPred == null)
      newPred = new StopWatch ();
    else
      newPred.resume ();
  }
  public void stopNewPred ()
  {
    newPred.pause ();
  }

  public void startWP ()
  {
    if (wp == null)
      wp = new StopWatch ();
    else
      wp.resume ();
  }

  public void stopWP ()
  {
    wp.pause ();
  }

  public void startTPQuery ()
  {
    if (tpQuery == null)
      tpQuery = new StopWatch ();
    else
      tpQuery.resume ();

    numQueries++;
  }

  public void stopTPQuery ()
  {
    tpQuery.pause ();
  }

  public void startTPAssert ()
  {
    if (tpAssert == null)
      tpAssert = new StopWatch ();
    else
      tpAssert.resume ();
  }

  public void stopTPAssert ()
  {
    tpAssert.pause ();
  }

  public void startTPGetAssm ()
  {
    if (tpIndices == null)
      tpIndices = new StopWatch ();
    else
      tpIndices.resume ();
  }

  public void stopTPGetAssm ()
  {
    tpIndices.pause ();
  }

  public void startTPWriteQuery ()
  {
    if (tpWriteQuery == null)
      tpWriteQuery = new StopWatch ();
    else
      tpWriteQuery.resume ();
  }

  public void stopTPWriteQuery ()
  {
    tpWriteQuery.pause ();
  }

  public void startAddrModel ()
  {
    if (addrModel == null)
      addrModel = new StopWatch ();
    else
      addrModel.resume ();
  }

  public void stopAddrModel ()
  {
    addrModel.pause ();
  }

  public void startRelatedPredicates ()
  {
    if (relatedPredicates == null)
      relatedPredicates = new StopWatch ();
    else
      relatedPredicates.resume ();
  }

  public void stopRelatedPredicates ()
  {
    relatedPredicates.pause ();
  }

  private StringBuffer appendnl (StringBuffer sb, String s)
  {
    sb.append (s);
    sb.append ("\n");
    return sb;
  }

  public String toString ()
  {
    StringBuffer sb = new StringBuffer ();
    appendnl (sb, "************************************");
    appendnl (sb, "STATISTICS");
    appendnl (sb, "Overall time: " + overall);
    appendnl (sb, "Parsing time: " + parse);
    appendnl (sb, "Refine time: " + refine);
    appendnl (sb, "Model compilation: " + compile);
    appendnl (sb, "Model-Checking time: " + mc);
    appendnl (sb, "New predicates time: " + newPred);
    appendnl (sb, "PredicateAbstraction time: " + predicateAbstraction);
    appendnl (sb, "\tRelated predicates: " + relatedPredicates);
    appendnl (sb, "WP time: " + wp);
    appendnl (sb, "Theorem-prover:");
    appendnl (sb, "\tQuery: " + tpQuery);
    appendnl (sb, "\t#Queries: " + numQueries);
    appendnl (sb, "\tWrite query: " + tpWriteQuery);
    appendnl (sb, "\tAssert: " + tpAssert);
    appendnl (sb, "\tGet assumptions: " + tpIndices);
    appendnl (sb, "Address model translation time: " + addrModel);
    //appendnl (sb, "Number of queries: " + queryTimes.size ());
    //appendnl (sb, "Query times: " + queryTimes.toString ());

    return sb.toString ();
  }
  
} // YasmStatistics
