package edu.toronto.cs.yasm.abstractor;


import edu.toronto.cs.yasm.refiner.*;
import edu.toronto.cs.yasm.pprogram.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.YasmStatistics;
import edu.toronto.cs.yasm.wp.*;

import edu.toronto.cs.tp.cvcl.CVCLUtil;


import java.util.*;


/**
 * Describe class CBJPredicateRefinerVisitor here.
 *
 *
 * Created: Wed Jul 20 16:17:15 2005
 *
 * @author <a href="mailto:owei@epoch.cs">Ou Wei</a>
 * @version 1.0
 */
public class CBJPredicateRefinerVisitor extends PredicateRefinerVisitor
{
  
  Map cache;

  List allPreds;

  boolean doInconsitent;
  boolean doSearch;

  public CBJPredicateRefinerVisitor (ExprFactory _fac)
  {
    super (_fac);
    doInconsitent = true;
    doSearch = true;
  }

  /**
   * doInconsitent -- if true compute inconsistent cubes
   * doSearch -- if true do abstraction search
   *
   * @param _fac an <code>ExprFactory</code> value
   * @param _doInconsitent a <code>boolean</code> value
   * @param _doSearch a <code>boolean</code> value
   */
  public CBJPredicateRefinerVisitor (ExprFactory _fac, boolean _doInconsitent,
      boolean _doSearch)
  {
    super (_fac);

    doInconsitent = _doInconsitent;
    doSearch = _doSearch;
  }

  //   public CBJPredicateRefiner (ExprFactory _fac, List _newPredicates)
  //   {
  //     super (_fac, _newPredicates);
  //   }


  public Object visitPProgram (PProgram p, Object o)
  {
    if (newPredicates.isEmpty () && !predicates.isEmpty ()) return p;

    // -- predicates is empty when this is the first run of this refiner
    if (predicates.isEmpty ())
      p.setInconsistent (fac.falseExpr ());

    cache = new IdentityHashMap ();

    allPreds = new ArrayList ();
    allPreds.addAll (predicates);
    allPreds.addAll (newPredicates);


    if (!allPreds.isEmpty () && doInconsitent)
    {

      TPCBJPredicateAbstractor cbj = 
        new TPCBJPredicateAbstractor (allPreds,
				      tp.exprFactory ().falseExpr (),
				      tp);
      cbj.recSearch (1);
      Expr inconsistent = cbj.getTExpr ();
      System.err.println ("INCONSISTENT: " + inconsistent);
      p.setInconsistent (inconsistent);
    }

    return super.visitPProgram (p, o);
  }

  private CBJRefinerInfo getRefinerInfo (PrllAsmtPStmt asmt)
  {
    CBJRefinerInfo info = (CBJRefinerInfo) asmt.getRefinerInfo ();
    if (info == null)
    {
      info = new CBJRefinerInfo ();
      asmt.setRefinerInfo (info);
    }
    return info;
  }

  public MemoryModel getRegularMemoryModel (PrllAsmtPStmt asmt)
  {
    return getRefinerInfo (asmt).memoryModel;
  }

  public CVCLMemoryModel getMemoryModel (PrllAsmtPStmt asmt)
  {
    return getRefinerInfo (asmt).cvclMemoryModel;
  }

  public WPComputer getWPComputer (PrllAsmtPStmt astmt)
  {
    return getRefinerInfo (astmt).wp;
  }

  /**
   * refine <code>visitPrllAsmt</code> method here.
   *
   * @param asmt a <code>PrllAsmtPStmt</code> value
   * @param o an <code>Object</code> value
   * @return a <code>PrllAsmtPStmt</code> value
   */
  public Object visitPrllAsmt (PrllAsmtPStmt asmt, Object o)
  {

    WPComputer wp = asmt.getWPComputer ();

    //List asmts = new ArrayList ();
    List asmts;
    int start;


    /**
     ** This is a bit confusing, but this is how it works:
     ** if doSearch is true, we go over all new and old predicates
     ** and compute the new abstraction of each statement. Therefore, 
     ** we start at the first predicate and an empty list

     ** If we do not do search, then we only need to compute new information
     ** for new predicates we have. Thus we start with previous computation
     ** which is stored in 'astm.getAsmts ()', and only process new predicates
     ** that start at the index 'predicates.size ()'
     **/
    if (doSearch)
    {
      start = 0;
      asmts = new ArrayList ();
    }
    else
    {
      asmts = asmt.getAsmts ();
      start = predicates.size ();
    }

    for (int i = start; i < allPreds.size (); i++)
    {
      //Expr target = converted.get (i);

      Expr wpExpr = wp.computeWP ((Expr) allPreds.get (i));

      assert wpExpr != null : "null wp for: " + asmt;


      Expr[] result = null;
      if (wpExpr.equals (allPreds.get (i)))
      {
        //System.err.println ("QUICKTEST1");
        result = 
          new Expr[] {fac.op (BoolOp.NOT).
            unaryApply ((Expr) allPreds.get (i)),
            (Expr) allPreds.get (i)};
      }
      else if (tp.quickQuery (wpExpr))
      {
        //System.err.println ("QUICKTEST2");
        result = new Expr[] 
        { fac.falseExpr (), fac.trueExpr () };
      }
      else if (tp.quickQuery 
	       (tp.exprFactory ().op (BoolOp.NOT).unaryApply (wpExpr)))
      {
        //System.err.println ("QUICKTEST3");
        result = new Expr[] 
        { fac.trueExpr (), fac.falseExpr () };
      }
      else
        result = (Expr[]) cache.get (wpExpr);

      if (result == null && doSearch)
      { 
        TPCBJPredicateAbstractor cbj
          = new TPCBJPredicateAbstractor (allPreds, 
					  wpExpr,
					  tp);
	
	if (stats != null)
	  stats.startPredicateAbstraction ();
        cbj.recSearch (1);
	if (stats != null)
	  stats.stopPredicateAbstraction ();
        result = new Expr[] {cbj.getFExpr (), cbj.getTExpr ()};
	cache.put (wpExpr, result);
      }
      else if (result == null && !doSearch)
        result = new Expr[] { fac.falseExpr (), fac.falseExpr () };
      else
        ; //System.err.println ("CACHE HIT");

      asmts.add 
        (new AsmtPStmt 
         ((Expr) allPreds.get (i), result [1], result [0]));
    }

    asmt.setAsmts (asmts);
    return asmt;
  }


  //   private boolean quickQuery (edu.toronto.cs.tp.cvcl.Expr expr)
  //   {

  //     int scope = vc.scopeLevel ();
  //     try 
  //       {
  // 	return vc.query (expr, false);
  //       }
  //     finally 
  //       {
  // 	vc.popto (scope);
  //       }

  //   }


  public class CBJRefinerInfo
  {
    MemoryModel memoryModel;
    CVCLMemoryModel cvclMemoryModel;
    WPComputer wp;

    // -- more stuff
  }


  
}
