
package edu.toronto.cs.yasm.eqsolver;

import java.util.*;

/**
 * Describe interface Equation here.
 *
 *
 * Created: Thu May 31 16:16:51 2007
 *
 * @author <a href="mailto:arie@pcawf.sei.cmu.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public interface Equation<V>
{
  /**
   * Returns the variable defined by this equation
   * 
   * @return variable defined by the equation
   */
  int getDefVar ();
  /**
   * Returns a list of all variables used by this equation
   *
   * @return a list of variables used by this equation 
   */
  List<Integer> getUseVars ();

  /**
   * Evaluates this equation under a given environment 
   *
   * @param env an environment that maps variables to values
   * @return the value of this equation under the enviroment
   */
  V eval (Map<Integer,V> env);


  /**
   * Returns true if the variable defined by the equation may change
   * when evaluated under the given environment.
   *
   * It is always safe for this method to return true
   *
   * @param env an environment
   * @return true if the variable defined by the equation may change
   * under env.
   */

  boolean maybeUnstable (Map<Integer, V> env);
  

  /**
   * Returns true if the variable defined by this equation may change
   * its value whenever the equation is reevaluated with new value of
   * the variable {@code var}. 
   *
   * The method must be implemented as efficient as possible. It is
   * always safe for this method to return true.
   *
   * @param newEnv the new enviroment
   * @param var the variable that was changed
   * @param oldV the old value of var
   * @param newV the new value of var, same as {@code newEnv[var]}.
   * @retun true if the value of defined variable may change
   */
  boolean mayChange (Map<Integer,V> newEnv, int var, V oldV, V newV);

  /**
   * Retunrs true when the current equation has reached its upper
   * limit, provided that the inputs will continue to increase.
   * 
   * @param env an  environment
   * @return true if defining variable reached its upper limit
   */
  boolean isUpwardStable (Map<Integer,V> env);
  
  /**
   * Retunrs true when the current equation has reached its lower
   * limit, provided that the inputs will continue to decrease.
   * 
   * @param env an  environment
   * @return true if defining variable reached its lower limit
   */
  boolean isDownwardStable (Map<Integer,V> newEnv);
}


