package edu.toronto.cs.yasm.instrumenter;

import java.io.*;
import java.util.*;

import antlr.*;
import antlr.collections.AST;
import antlr.debug.misc.ASTFrame;

import edu.toronto.cs.cparser.*;
import edu.toronto.cs.cparser.block.*;

public class SimpleInstrumenter
{
  
  
  static HashMap map = new HashMap();
  
  private SimpleInstrumenter () {}

  // -- show a tree representation of an AST
  public static void drawFrame(String name, AST ptree)
  {
    ASTFrame frame = new ASTFrame(name, ptree);
    frame.setVisible(true);
  }

  public static String readFileIntoString(String f) throws Exception{
    BufferedReader bufReader = new BufferedReader(new FileReader (f));
    String fileText = ""; // this variable will store unparsed instructions
    String line = bufReader.readLine();
    while(line != null){
      fileText = fileText + "\n" + line;
      line = bufReader.readLine();
    }
    return fileText;
  }
  
  /*
   * This method reads inctructions file in new format
   */
  
  public static void readInstructions2(String instrFile) throws Exception
  {
        
    StringTokenizer tokenizer = new StringTokenizer(readFileIntoString(instrFile));
    int statementNumber = -1; 
    String[] statements = {"","","","",""}; 
    String fName = ""; 
    String temp;
    int stack;

    // Start reading through the fileText and parsing instructions

    while(tokenizer.hasMoreElements()){
      temp = tokenizer.nextToken(); // reads the first token which is functName.location
      
      if(temp.indexOf(".entry") != -1){statementNumber = 2;}
      else if(temp.indexOf(".exit") != -1){statementNumber = 3;}
      else if(temp.indexOf(".call") != -1){statementNumber = 0;}
      else if(temp.indexOf(".return") != -1){statementNumber = 1;}
      else if(temp.indexOf("state") != -1){statementNumber = 4;} 
      else{throw new Exception("invalid instructions format. token is--" + temp + "--, not functName.location");} 
      
      if(statementNumber != 4) fName = temp.substring(0,temp.indexOf(".")); //read function name if there is one 
      temp = tokenizer.nextToken(); // skips {
      
      stack = 1;
      while(stack != 0){
         temp = tokenizer.nextToken();  
	 if(temp.equals("{")){stack++;}
	 else if(temp.equals("}")){stack--;}
         if(stack !=0){statements[statementNumber] = statements[statementNumber] + " " + temp;} 
      }
      if(statementNumber == 4){  // if we are now working with state
        map.put("state", statements[4]);
      }
      // if we are not then the function name must be present. Check if it is already in the map  
      else{
        String[] tempArray = {"---","---","---","---"};
	if(map.containsKey(fName)){ 
	  tempArray = (String[])map.get(fName); 
	}
        tempArray[statementNumber] = statements[statementNumber];
        map.put(fName, tempArray);
      }
      statements[statementNumber] = ""; 
    } 
  }
  
  


  

  public static void main(String[] args) throws Exception{
    // for (int i = 0; i < args.length; i++)
    //{
      
      String programName = args[0];
      Reader reader = new FileReader (programName);

      // Read the instructions file if it is given
      if(args.length > 1 ){
	String instructions = args[1];
        readInstructions2(instructions);
      
   

      }
      
      GnuCLexer lexer =
        new GnuCLexer ( reader );
      lexer.setTokenObjectClass("edu.toronto.cs.cparser.CToken");

      // Parse the input expression.
      GnuCParser parser = new GnuCParser ( lexer );

      // set AST node type to TNode or get nasty cast class errors
      parser.setASTNodeType (TNode.class.getName());
      TNode.setTokenVocabulary ("edu.toronto.cs.cparser.GNUCTokenTypes");

      parser.translationUnit ();

			TNode root = new TNode ();
			root.setFirstChild (parser.getAST ());
     //	drawFrame ("Program AST", root);
     //    if (2 == 2)
     //	return;
      
      //System.exit (1);
	//System.out.println (parser.getAST ());

	//System.out.println (parser.getAST ().toStringTree ());
        GnuCInstrEmitter treeParser =
		new GnuCInstrEmitter (new PreprocessorInfoChannel (), map);
	treeParser.setASTNodeClass (TNode.class.getName ());

	// treeParser.setPrintStream (System.out);

	
	treeParser.translationUnit (parser.getAST ());
      // AST t = parser.getAST ();

      //System.out.println (parser.symbolTable.toString ());

      //if (t != null)
      //  drawFrame ("CIL Program AST", t);
      //}
      }
      
  }
  


