package edu.toronto.cs.yasm.nontermination;

import edu.toronto.cs.yasm.pcfa.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.yasm.reachability.ReachabilityStructure;


import java.util.*;

/**
 * Describe class NonTermination  here.
 * <p>
 * check EG true
 *
 * Created: Wed Dec  9 11:12:35 2005
 *
 * @author <a href="mailto:owei@eon.cs">Ou Wei</a>
 * @version 1.0
 */
public class NonTerminationChecker 
{

  boolean DEBUG = false;
  

  ReachabilityStructure model;
  
  Collection workList;
  Collection sumWorkList;

  Collection oldWorkList;
  Collection oldSumWorkList;

  
  /**
   * NodeSet for reachability phase of the analysis
   *
   */
  NodeSet nodes;
  
  /**
   * NodeSet for summary computation
   *
   */
  NodeSet sumNodes;


  /**
   * Creates a new <code>NonTerminationChecker</code> instance.
   *
   * @param _model a <code>ReachabilityStructure</code> value
   */
  public NonTerminationChecker (ReachabilityStructure _model) 
  {
    workList = new LinkedList ();
    sumWorkList = new LinkedList ();
    model = _model;
  }


  public ReachabilityStructure getModel ()
  {
    return model;
  }


  public void doNonTermination ()
  {
    initialize ();

    do
      {
	// -- do a step in nontermination computation
	doNonTerminationStep ();


	// -- check if there is more work to do and 
	// -- if so, update nodes
	if (!workList.isEmpty ())
	  {
	    model.setNodeSet (nodes);
	    nodes = nodes.createNext ();
	  }
	
      }
    while (! workList.isEmpty ());

    model.setNodeSet (nodes);
    model.setSumNodeSet (sumNodes);
  }

  /**
   * Describe <code>initialize</code> method here.
   *
   */
  private void initialize ()
  {
    // -- set nodes and sumNodes
    //nodes = model.getNodeSet ().createNext ();
    //    sumNodes = model.getSumNodeSet ().createNext ();
    sumNodes = model.getSumNodeSet ();
    nodes = model.getNodeSet ();

    // -- set each node with "TRUE" except for "END" node which has
    // -- been set "FALSE" in YasrApp.java

    // -- create initial worklist, add all edges

    for (Iterator it = model.getEdgeSet ().getSrcMap ().values ().iterator (); 
	 it.hasNext ();)
      {

	Collection edges = (Collection) it.next ();

	for (Iterator itEdge = edges.iterator (); itEdge.hasNext ();)
	  {
	    Edge edge = (Edge) itEdge.next ();

	    //	    System.err.println ("edge : " + edge);
	    
	    if (edge instanceof ReturnEdge)
	      {
		setNodeFalse (edge.getSrc ());
		continue;
		
	      }
	    else
	      {
		setNodeTrue (edge.getSrc ());
	      }
	    
	    // 	    if (edge instanceof MvRelnHyperEdge)
	    // 	      {
	    // 		Node.NodeId[] allDst;

	    // 		allDst = edge.getAllDst ();
	    // 		for (int i = 0; i < allDst.length; i++)
	    // 		  {
	    // 		    setNodeTrue (allDst [i]);
	    // 		  }
	    // 	      }
	    // 	    else
	    // 	      {
	    // 		setNodeTrue (edge.getDst ());
	    // 	      }
	  }
    
	
	workList.addAll (edges);    
      }    

    // -- at this point nodes is ready to be the initial step 
    // -- of the computation and we need a fresh copy to store results
    nodes = model.getNodeSet ().createNext ();
    
  }
  
          
  public void doNonTerminationStep ()
  {
   
    // -- store old worklists
    oldWorkList = workList;

    
    // -- create new ones
    workList = new LinkedList ();
    
    System.err.println (" ***** do One EX computation *****");

    computeEX ();
    
  }


  private void computeEX ()
  {
    
    // -- process work list
    for (Iterator it = oldWorkList.iterator (); it.hasNext (); )
      {
	Edge edge = (Edge) it.next ();

	// -- skip ReturnEdgee
	if (edge instanceof ReturnEdge)
	  {
	    continue;
	  }


	// -- on a CallEdge or FunctionSummaryEdge we start a new
	// -- summary computation, we start from the exit node of the
	// -- function based on the information on the corresponding
	// -- ReturnEdge
	//
	// -- after the summary computation is finished, we compute EX
	// -- on the FunctionSummaryEdge and the CallEdge 
	//
	// -- see the comments for "visitFunctionCallPrologue" in
	// -- "LocalPProgramReachabilityCompilerVisitor.java"

	if (edge instanceof CallEdge || edge instanceof FunctionSummaryEdge)
	  {

	    CallEdge callEdge;
	    FunctionSummaryEdge funcSumEdge;
	    ReturnEdge returnEdge;
	    
	    
	    if (edge instanceof CallEdge)
	      {
		callEdge = (CallEdge) edge;
		funcSumEdge = callEdge.getFunSummary ();
		returnEdge = callEdge.getReturnEdge ();
	      }
	    else // FunctionSummaryEdge
	      {
		funcSumEdge = (FunctionSummaryEdge)edge;
		callEdge = funcSumEdge.getCallEdge ();
		returnEdge = callEdge.getReturnEdge ();
	      }
	    
	    
	    
	    // -- extract a relation from the edge
	    MvRelation reln = returnEdge.getRelation ();
	    // -- propogate dst labeling over this relation
	    MvSet v = reln.bwdImage 
	      (nodes.get (returnEdge.getDst ()).getValue ());

	    // -- convert the set into a subset of the identity relation
	    v = setToReln (v, returnEdge.getSrcIdReln ());
	    
	    // -- update sumNodes and add edge to summary worklist
	    Node exitNode = sumNodes.get (returnEdge.getSrc ());
	    if (exitNode != null)
	      v = v.or (exitNode.getValue ());
	    
	    // -- check if anything has changed
	    if (exitNode == null || ! v.equals (exitNode.getValue ()))
	      {
		// -- update summary nodes 
		sumNodes.add (returnEdge.getSrc (), v);

		// -- add edges to exitNode to sumWorkList
		sumWorkList.addAll
		  (model.getEdgeSet ().edgesTo (returnEdge.getSrc ()));


		model.setSumNodeSet (sumNodes);
		sumNodes = sumNodes.createNext ();


		// -- computer function summary
		System.err.println (" ***** start Computing Summary  : " +
				    funcSumEdge + " *****");
		
		computeSummary ();

		System.err.println (" ***** end Computing Summary *****");

	      }	 

	    // -- now compute EX on the CallEdge and FunctionSummaryEdge

	    
	    bwdImage ( callEdge,
		       funcSumEdge,
		       model.getNodeSet (), 
		       nodes, 
		       workList);

	  }
	else
	  {
	    // -- for all other edges, just compute the bwdImage 
	    bwdImage (edge, model.getNodeSet (), nodes, workList);
	  }
	
      }

  }

  
  private void computeSummary ()
  {
    do
      {
	// -- store old worklists
	oldSumWorkList = sumWorkList;

	sumWorkList = new LinkedList ();

	
	while (!oldSumWorkList.isEmpty ())
	  {
	    Edge edge = (Edge) ( (LinkedList)oldSumWorkList).remove ();

	    // -- for return edge we start a new summary computation 
	    // -- just like before
	    if (edge instanceof ReturnEdge)
	      {

		System.err.println (" reach Return Edge : " + edge);

		// -- extract a relation from the edge
		MvRelation reln = ((ReturnEdge)edge).getRelation ();
		// -- propogate dst labeling over this relation
		MvSet v = reln.bwdImage 
		  (sumNodes.get (edge.getDst ()).getValue ());

		// -- convert the set into a subset of the identity relation
		v = setToReln (extractPreStates (v), 
			       ((ReturnEdge)edge).getSrcIdReln ());

		Node exitNode = sumNodes.get (edge.getSrc ());
		if (exitNode != null)
		  {
		    v = v.or (exitNode.getValue ());		
		  }
	    
		if (exitNode == null || ! v.equals (exitNode.getValue ()))
		  {
		    sumNodes.add (edge.getSrc (), v);
		    sumWorkList.addAll 
		      (model.getEdgeSet ().edgesTo (edge.getSrc ()));
		  }
	      }
	    // -- reaching a call edge indicates that a summary computation
	    // -- has converged and we should instantiate new summary edges
	    else if (edge instanceof CallEdge)
	      {
		System.err.println (" reach CallEdge : " + edge);

		// -- time to update summaries

		// -- extract current labeling of function entry node
		MvSet summary = sumNodes.get (edge.getDst ()).getValue ();

		// -- convert a summary to a relation (i.e. expression over 
		// -- pre- and post- variables 
		MvSet sumReln = summaryToRelation (summary);

	    
		// 		System.err.println ("Got a new sumReln: ");
		// 		System.err.println (new Node (new Node.NodeId (0),
		// 					      sumReln));
		// 		System.err.println ("Done with new summary");
	    
		// -- update the body summary edge
		// -- and get a new function summary
		SummaryEdge funSummary = 
		  ((CallEdge) edge).updateBodySummary (sumReln,
						       model.getPostToAuxMap (),						       model.getAuxToPostMap ());
	    
		// -- add function summary to the worklist
		sumWorkList.add (funSummary);
	      }
	    else
	      bwdImage (edge, model.getSumNodeSet (), sumNodes, sumWorkList);
	  } 

	if (!sumWorkList.isEmpty ())
	  {
	    model.setSumNodeSet (sumNodes);
	    sumNodes = sumNodes.createNext ();
	  }    
      }
    while (! sumWorkList.isEmpty ());
    

  }
  


  private MvSet summaryToRelation (MvSet v)
  {
    return v.renameArgs (model.getAuxToPostMap ());
  }
  
  private MvSet extractPreStates (MvSet v)
  {
    return v.existAbstract (model.getAuxCube ());
  }
  
  private MvSet setToReln (MvSet set, MvSet idReln)
  {
    //     return set.renameArgs (model.getPreToAuxMap ()).and 
    //       (idReln.renameArgs (model.getPostToAuxMap ()));


    //return idReln.renameArgs (model.getPostToAuxMap ());

    return idReln.and (set).renameArgs (model.getPostToAuxMap ());

  }
  
  private void bwdImage (Edge edge, NodeSet oldNodes, 
			 NodeSet nodes, 
			 Collection _workList)
  {
    System.err.println ("bwdImage over edge: ( " + edge.getClass () 
			+ " ) " + edge);

    if (DEBUG)
      {    
	System.err.println (" *** before computation *** ");

	System.err.println ("look at Node : " + edge.getSrc ());
	System.err.println (nodes.get (edge.getSrc ()));
      }
	
    Node.NodeId[] allDst;

    allDst = edge.getAllDst ();

    if (DEBUG)
      {
	for (int i = 0; i < allDst.length; i++)
	  {
	    System.err.println ("look at Node : " + allDst [i]);
	    System.err.println (nodes.get (allDst [i]));
	  }
      }
	

    if (edge.bwdImage (oldNodes, nodes, false))
      {
	_workList.addAll (model.getEdgeSet ().edgesTo (edge.getSrc ()));
      }  

    if (DEBUG)
      {    

	System.err.println (" *** after computation *** ");
	    
	System.err.println ("look at Node : " + edge.getSrc ());
	System.err.println (nodes.get (edge.getSrc ()));
      }
	

  }
  

  /**
   * compute EX on <code>callEdge</code> and <code>funcSumEdge</code>
   * at the same time.
   *
   * @param callEdge a <code>CallEdge</code> value
   * @param funcSumEdge a <code>FunctionSummaryEdge</code> value
   * @param oldNodes a <code>NodeSet</code> value
   * @param nodes a <code>NodeSet</code> value
   * @param workList a <code>Collection</code> value
   */
  private void bwdImage (CallEdge callEdge, 
			 FunctionSummaryEdge funcSumEdge,
			 NodeSet oldNodes, 
			 NodeSet newNodes,
			 Collection _workList)
  {
    System.err.println ("bwdImage over  CallEdge : " + callEdge 
			+ "   and FunctionSummaryEdge : " + funcSumEdge);

    if (DEBUG)
      {    

        System.err.println (" *** before computation *** ");

        System.err.println ("look at Node : " + callEdge.getSrc ());
        System.err.println (nodes.get (callEdge.getSrc ()));

        System.err.println ("look at Node : " + callEdge.getDst ());
        System.err.println (nodes.get (callEdge.getDst ()));

        System.err.println ("look at Node : " + funcSumEdge.getDst ());
        System.err.println (nodes.get (funcSumEdge.getDst ()));
	
      }
	

    // -- similar to bwdImage () in MvRelnEdge.java

    // XXX problem, what if getDst is null?!
    MvSet res = null;
    
    Node dstNode;

    // -- get the value that callEdge points to
    dstNode = oldNodes.get (callEdge.getDst ());    

    // -- compute backward image over the call edge
    if (dstNode != null)
      res = callEdge.getRelation ().bwdImage (dstNode.getValue ());

    // -- get the value that funcSumEdge points to
    dstNode = oldNodes.get (funcSumEdge.getDst ());
    
    // -- compute backward image over function summary and add it
    // -- to the previous result
    if (dstNode != null)
      res = 
	res.or (funcSumEdge.getRelation ().bwdImage (dstNode.getValue ()));


    // -- find the node from which callEdge is pointing from
    // -- this is the same as the source of the funcSummaryEdge
    Node srcNode = newNodes.get (callEdge.getSrc ());
    
    // -- check if what we have computed is different from what was 
    // -- there before
    // -- i.e., res is non-null, and srcNode is different
    if (res != null && 
	(srcNode == null || !res.equals (srcNode.getValue ())))
      {
	// -- update the srcNode value
	newNodes.add (callEdge.getSrc (), res);
	
	// -- add all edges to this value to the workList
	_workList.addAll 
	  (model.getEdgeSet ().edgesTo (callEdge.getSrc ()));	
      }    

    if (DEBUG)
      {
	System.err.println (" *** after computation *** ");
	    
	System.err.println ("look at Node : " + callEdge.getSrc ());
	System.err.println (nodes.get (callEdge.getSrc ()));
      }
	

  }


  private void setNodeTrue (Node.NodeId _nodeId)
  {
    //    System.out.println ("Trying to set node: " + _nodeId);
    Node tmpNode = nodes.get (_nodeId);

    if ( tmpNode  == null)
      {    
	//	    System.out.println ("Set node to true: " + _nodeId);
	    
	nodes.add (new Node (_nodeId, 
			     model.getMvSetFactory ().top ()));
      }
  }
  

  private void setNodeFalse (Node.NodeId _nodeId)
  {
    //    System.out.println ("Trying to set node: " + _nodeId);
    Node tmpNode = nodes.get (_nodeId);

    if ( tmpNode  == null)
      {    
	//	    System.out.println ("Set node to false: " + _nodeId);
	    
	nodes.add (new Node (_nodeId, 
			     model.getMvSetFactory ().bot ()));
      }
  }



}
