package edu.toronto.cs.yasm.pcfa;

import java.util.*;


/**
 * Describe class EdgeSet here.
 *
 *
 * Created: Wed Jun  8 13:45:57 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class EdgeSet 
{

  Map dstMap;
  Map srcMap;
  
  /**
   * Creates a new <code>EdgeSet</code> instance.
   *
   */
  public EdgeSet () 
  {
    dstMap = new HashMap ();
    srcMap = new HashMap ();
  }

  
  /**
   * Iterator over edges in this set
   *
   * @return an <code>Iterator</code> value
   */
  public Iterator iterator ()
  {
    return null;
  }

  /**
   * Returns a subset of all edges whose src is id
   *
   * @param id a <code>Node.NodeId</code> value
   * @return an <code>Collection</code> value
   */
  public Collection edgesTo (Node.NodeId id)
  {
    return getEdges (dstMap, id);    
  }

  /**
   * Returns a subset of all edges whose dst contains id
   *
   * @param id a <code>Node.NodeId</code> value
   * @return an <code>Collection</code> value
   */
  public Collection edgesFrom (Node.NodeId id)
  {
    return getEdges (srcMap, id);
  }

  private Collection getEdges (Map map, Node.NodeId id)
  {
    Collection edges = (Collection) map.get (id);
    return edges == null ? Collections.emptyList () : edges;
  }
  
  
  public Collection edges (Node.NodeId fromId, Node.NodeId toId)
  {
    Collection out = new LinkedList ();
    for (Iterator it = edgesFrom (fromId).iterator (); it.hasNext ();)
      {
	Edge edge = (Edge) it.next ();
	if (edge.hasDst (toId)) out.add (edge);
      }
    return out;
    
  }

  public Edge edge (Node.NodeId fromId, Node.NodeId toId)
  {
    for (Iterator it = edgesFrom (fromId).iterator (); it.hasNext ();)
      {
	Edge edge = (Edge) it.next ();
	if (edge.hasDst (toId)) return edge;
      }
    return null;
  }
  
  
  
  /**
   * Adds an edge
   *
   * @param edge an <code>Edge</code> value
   */
  public void add (Edge edge)
  {
    addEdge (srcMap, edge.getSrc (), edge);
    if (edge instanceof HyperEdge)
      {
	Node.NodeId[] dsts = edge.getAllDst ();
	for (int i = 0; i < dsts.length; i++)
	  addEdge (dstMap, dsts [i], edge);
      }
    else
      addEdge (dstMap, edge.getDst (), edge);
  }

  /**
   * Returns the size of this set
   *
   * @return an <code>int</code> value
   */
  public int size ()
  {
    return 0;
  }
  

  /**
   * return <code>srcMap</code>
   *
   * @return a <code>Map</code> value
   */
  public Map getSrcMap ()
  {
    return srcMap;
    
  }
  
  /**
   * return <code>dstMap</code>
   *
   * @return a <code>Map</code> value
   */
  public Map getDstMap ()
  {
    return dstMap;
    
  }


  private void addEdge (Map map, Node.NodeId id, Edge edge)
  {
    Collection edges = (Collection) map.get (id);
    if (edges == null)
      {
	edges = new LinkedList ();
	map.put (id, edges);
      }
    
    edges.add (edge);
  }
  
  
  
  

}
