package edu.toronto.cs.yasm.pcfa;

import edu.toronto.cs.mvset.*;

/* needed only for dumpMvSet */
import java.util.Iterator;
import java.util.Arrays;


/**
 * Describe class MvRelnHyperEdge here.
 *
 *
 * Created: Wed Jun  8 15:02:58 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class MvRelnHyperEdge extends HyperEdge 
{

  MvRelation[] relns;

  // -- if true, bwdImage computation does AX to increase precision
  boolean doAX;

  public MvRelnHyperEdge (Node.NodeId _src, Node.NodeId[] _dsts,
			  MvRelation[] _relns, boolean _doAX)
  {
    super (_src, _dsts);
    relns = _relns;
    doAX = _doAX;
  }
  
  /**
   * Creates a new <code>MvRelnHyperEdge</code> instance.
   *
   */
  public MvRelnHyperEdge (Node.NodeId _src, Node.NodeId[] _dsts, 
			  MvRelation[] _relns) 
  {
    this (_src, _dsts, _relns, true);
  }

  public boolean bwdImage (NodeSet oldNodes, 
			   NodeSet newNodes,
			   boolean isDisjoin)
  {
    boolean changed = false;
    MvSet res = null;
    
    MvSet tmp;
    
    for (int i = 0; i < dsts.length; i++)
      {
	Node dstNode = oldNodes.get (dsts [i]);
	// -- skip null dst nodes
	if (dstNode == null)
	    continue;
	
	
	tmp = relns [i].bwdImage (dstNode.getValue ());

//  	System.err.println ("doing bwdImage of: ");
//  	dumpMvSet (dstNode.getValue ());
	
//  	System.err.println ("result is: ");
//  	dumpMvSet (tmp);

	if (res != null) 
	  res = res.or (tmp);
	else
	  res = tmp;
      }

    assert res != null : "res is null, should not happen";
    
    if (doAX)
      {
	// -- compute AX
	MvSet axRes = null;
	for (int i = 0; i < dsts.length; i++)
	  {
	    Node dstNode = oldNodes.get (dsts [i]);
	    MvSet dstVal = null;
	    if (dstNode == null) 
	      dstVal = res.getFactory ().bot ();
	    else
	      dstVal = dstNode.getValue ();
	    
	    tmp = relns [i].dualBwdImage (dstVal);
	    if (axRes != null)
	      axRes = axRes.and (tmp);
	    else
	      axRes = tmp;
	  }
    
	assert axRes != null : "axRes is null, should not happen";

//      System.err.println ("Done AX Computation, got: ");
//      dumpMvSet (axRes);
//      System.err.println ("END Done AX");
    

     //res = res.or (axRes.geq (axRes.getFactory ().infoTop ()));

	axRes = axRes.geq (axRes.getFactory ().infoTop ()).
	  and (axRes.getFactory ().infoTop ());

	res = res.or (axRes);
      }
   
    Node src = newNodes.get (getSrc ());
    if (src != null)
      {
	if (isDisjoin)
	  {
	    res = res.or (src.getValue ());
	  }

	changed = !res.equals (src.getValue ());
      }
    else
      changed = true;
    
    if (changed)
      newNodes.add (getSrc (), res);

    return changed;
  }

    


  public boolean fwdImage (NodeSet oldNodes, NodeSet newNodes)
  {
    boolean globalChanged = false;
    
    for (int i = 0; i < dsts.length; i++)
      {
	boolean changed = false;
	MvSet res = relns [i].fwdImage (oldNodes.get (getSrc ()).getValue ());
	Node dst = newNodes.get (dsts [i]);
	if (dst != null)
	  {
	    res = res.or (dst.getValue ());
	    changed = !res.equals (dst.getValue ());
	  }
	else
	  changed = true;

	if (changed)
	  newNodes.add (dsts [i], res);

	globalChanged = globalChanged || changed;
      }

    return globalChanged;
  }
  

  private static void dumpMvSet (MvSet mvSet)
  {
    System.err.println ("mvSet of size: " + mvSet.size ());
    for (Iterator it = mvSet.cubeIterator (); it.hasNext ();)
      System.err.println (Arrays.asList ((Object[])it.next ()).toString ());
  }

  

}
