package edu.toronto.cs.yasm.pgenerator;

import java.util.*;

import edu.toronto.cs.tp.ITheoremProver;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.expr.BoolOp;
import edu.toronto.cs.yasm.pprogram.PProgram;
import edu.toronto.cs.yasm.pprogram.PrllAsmtPStmt;
import edu.toronto.cs.yasm.pprogram.AsmtPStmt;
import edu.toronto.cs.yasm.wp.*;
import edu.toronto.cs.yasm.abstractor.Refiner;



public class BackEndGenerator extends BasePredicateGenerator
{

  boolean worked = false;
  public BackEndGenerator (PProgram _pProgram,
      PrllAsmtPStmt _stmt,
      Refiner _refiner,
      WPComputer _wp,
      List<Expr> _pStatePred, 
      List<Expr> _cStatePred)

  {
    super (_pProgram, _stmt,_refiner, _wp,_pStatePred, _cStatePred);    
  }



  public boolean find ()
  {
    int scope = tp.stackLevel ();
    try
    {
      // -- push current scope and assume constraints of 
      // -- the parent state
      tp.push ();
      tp.assertFormula (pStatePred);
      // -- make sure we are in a consistent state
      if (isInconsistent ())
        return true;

      // -- refine the statement, as a side-effect worked
      // -- is set to true if the refinement process worked
      refineStatement ();
      return worked;
    }
    finally
    {
      tp.popto (scope);
    }
  }

  boolean isInconsistent ()
  {
    int scope = tp.stackLevel ();
    try
    {
      // 	System.err.println ("Checking inconsistency of: " + 
      // 			    vc.andExpr (pStatePred));
      tp.push ();
      // -- if everything is consistent nothing to do
      if (!tp.query (tp.exprFactory ().falseExpr ())) 
        return false;

      Expr newInconsistent = 
	tp.exprFactory ().op (BoolOp.AND).naryApply 
	(tp.getAssumptionsUsed ());
      
      System.err.println ("New inconsistent constraint: " + 
          newInconsistent);

      // -- 
      PProgram prog = getPProgram ();
      ExprFactory fac = prog.getInconsistent ().getFactory ();
      prog.
        setInconsistent (fac.op (BoolOp.OR).
			 binApply (prog.getInconsistent (), newInconsistent));
      return true;
    }
    finally
    {
      tp.popto (scope);
    }
  }

  void refineStatement ()
  {
    for (Iterator it = stmt.getAsmts ().iterator (); it.hasNext ();)
    {
      AsmtPStmt asmt = (AsmtPStmt) it.next ();
      // -- get the value of current variable in the child state
      boolean childVal = getChildVal (asmt.getVar ());


      // -- skip variables for which the next-state change is definite
      if (!childVal && tp.quickQuery (asmt.getFalseCond ()))
      {
	System.out.println ("BACKEND falseCond: " + asmt.getFalseCond ());      
        continue;
      }
      else if (childVal && tp.quickQuery (asmt.getTrueCond ()))
        continue;

      // -- if we got here then we can refine current assignment
      refineAsmt (asmt);
    }
  }


  boolean getChildVal (Expr expr)
  {
    for (Expr child : cStatePred)
      {
	if (child == expr) return true;
	else if (child.op () == BoolOp.NOT &&
		 child.arg (0) == expr) return false;
      }
    
    // XXX  no child value known, assume true
    return true;
    
    //assert false : "Was not able to find child expression for " + expr;
    //throw new RuntimeException ("Should never be here");
  }
  void refineAsmt (AsmtPStmt asmt)
  {
    int scope = tp.stackLevel ();
    edu.toronto.cs.expr.Expr wpExpr = wp.computeWP (asmt.getVar ());
    if (NondetOp.hasNondet (wpExpr))
      {
	System.err.println ("CANNOT REFINE NON-DETERMINISTIC ASSIGNMENT");
	return;
      }
    
    try
    {
      tp.push ();

      Boolean result = tp.checkTruth (computeWP (wpExpr));

      if (result != null)
      {
        worked = true;
	List<Expr> assumptionsUsed = tp.getAssumptionsUsed ();
	Expr assumptions;
	if (assumptionsUsed.size () == 1)
		assumptions = assumptionsUsed.get (0);
	else	
        	assumptions = tp.exprFactory ().op (BoolOp.AND).
		  naryApply (tp.getAssumptionsUsed ());

        System.err.println ("Got new assumptions: " + assumptions + 
			    " for " + (result == Boolean.TRUE ? "" : "!") + 
			    asmt.getVar ());

        ExprFactory fac = tp.exprFactory ();

        if (result == Boolean.TRUE)
          asmt.setTrueCond 
            (fac.op (BoolOp.OR).
             binApply (asmt.getTrueCond (), assumptions));
        else
          asmt.setFalseCond 
            (fac.op (BoolOp.OR).
             binApply (asmt.getFalseCond (), assumptions));
      }
    }
    finally
    {
      tp.popto (scope);
    }
  }


  public List<Expr> getNewPreds ()
  {
    return new ArrayList<Expr> ();
  }

  public void _backEndRefine ()
  {
    int scope = tp.stackLevel ();


    System.err.println ("ASSERTING: " + pStatePred);

    tp.push ();
    tp.assertFormula (pStatePred);

    int scope2 = tp.stackLevel ();

    for (Expr e : cStatePred)
      {
	for (int j = 0; j < 2; j++)
	  {
	    if (j == 1) e = flipExpr (e);
	    tp.push ();
      // XXX I THINK it only expects the pos wp here
	    if (!tp.query (ChoiceOp.lgetT (computeWP (e))))
	      {
		tp.popto (scope2);
		continue;
	      }
	    List<Expr> assumptions = tp.getAssumptionsUsed ();
	    Expr assumption = 
	      tp.exprFactory ().op (BoolOp.AND).naryApply (assumptions);
	    
	    improveAbstraction (assumption, e);
	    tp.popto (scope2);
	    break;
	  }
      }

    tp.popto (scope);
  } 

  private Expr flipExpr (Expr expr)
  {
    if (expr.op () == BoolOp.NOT) return expr.arg (0);
    else
      return tp.exprFactory ().op (BoolOp.NOT).unaryApply (expr);
  }

  private void improveAbstraction (Expr assumption, Expr e)
  {

    boolean found = false;
    boolean truePred;
    if (e.op () == BoolOp.NOT)
    {
      truePred = false;
      e = e.arg (0);
    }
    else
      truePred = true;

    List asmts = stmt.getAsmts ();
    System.out.println ("Adding " + assumption + " to " + 
        (truePred ? "" : "!") + e);

    for (int i = 0; i < asmts.size (); i++)
    {
      AsmtPStmt predAsmt = (AsmtPStmt) asmts.get (i);
      
      
      if (e == predAsmt.getVar ())
	{
	  if (truePred && !tp.implies (assumption, predAsmt.getTrueCond ()))
	    {
	      predAsmt.setTrueCond 
		(tp.exprFactory ().op (BoolOp.OR).binApply 
		 (predAsmt.getTrueCond (), assumption));
	      worked = true;
	    }
	  else if (!truePred && 
		   !tp.implies (assumption, predAsmt.getFalseCond ()))
	    {
	      predAsmt.setFalseCond 
		(tp.exprFactory ().op (BoolOp.OR).binApply
		 (predAsmt.getFalseCond (), assumption));
	      worked = true;
	    }
	  found = true;
	  break;
	}
    }

    System.err.println ("Updated parallel assignment to");
    stmt.print (new java.io.PrintWriter (System.err, true));

    if (!found)
      throw new RuntimeException ("BackEndGenerator.java Error 2");
  }

}
