package edu.toronto.cs.yasm.pgenerator;

import edu.toronto.cs.tp.ITheoremProver;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.wp.WPComputer;
import edu.toronto.cs.yasm.pprogram.PrllAsmtPStmt;
import edu.toronto.cs.yasm.pprogram.PProgram;
import edu.toronto.cs.yasm.abstractor.Refiner;


import java.util.*;


public abstract class BasePredicateGenerator implements PredicateGenerator
{
  protected PrllAsmtPStmt stmt;
  protected ITheoremProver tp;
  protected List<Expr> existingPred;

 
  
  // -- symbolic execution that we want to get rid of
  //protected Map symbolicMap;
  protected WPComputer wp;
  
  // -- parent state
  protected List<Expr> pStatePred;
  // -- child state
  protected List<Expr> cStatePred;


  protected List<Expr> newPreds;

  
  protected PProgram pProgram;

  
  /**
   * The base class for predicate generators.
   *
   * @param _pProgram a {@code PProgram} for which predicates are
   * being generateed.
   * @param _stmt an assignment from which predicates are inferred
   * @param _refiner a <code>Refiner</code> used to refine
   * the program
   * @param _wp a <code>WPComputer</code> for the assignment
   * @param _pStatePred a list of predicates true before the assignement
   * @param _cStatePred a list of predicates true after the assignment
   */
  public BasePredicateGenerator (PProgram _pProgram, 
				 PrllAsmtPStmt _stmt,
				 Refiner _refiner,
				 WPComputer _wp,
				 List<Expr> _pStatePred, 
				 List<Expr> _cStatePred)
				 
  {
    pProgram = _pProgram;
    stmt = _stmt;
    tp = _refiner.getTP ();
    existingPred = _refiner.getPredicates ();
    wp = _wp;
    pStatePred = _pStatePred;
    cStatePred = _cStatePred;

    newPreds = new ArrayList<Expr> ();

    // Class invariants:

    // assert wp != null : "WP is null";
  }




  /**
   * Describe <code>getPProgram</code> method here.
   *
   * @return a <code>PProgram</code> value
   */
  public PProgram getPProgram ()
  {
    return pProgram;
  }
  
  /**
   * Returns {@code null} if {@code expr} is already known by a {@code
   * PredicateRefiner}, and a negation free version of {@code expr}
   * otherwise.
   *
   * @param expr an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public Expr hasPredicate (Expr expr)
  {
    // -- get rid of negation if there is one
    if (expr.op () == BoolOp.NOT)
      expr = expr.arg (0);

    return existingPred.contains (expr) ? null : expr;
  }
  
  
  /**
   * Computes the weakest precondition of {@code Expr} over the
   * assignment of this predicate generator.
   *
   * @param expr an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  protected Expr computeWP (Expr expr)
  {
    if (NondetOp.hasNondet (expr)) return null;
    return wp.computeWP (expr);
  }

  public List<Expr> getNewPreds ()
  {
    return newPreds;
  }
 
}
