package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.expr.*;
import java.io.PrintWriter;

/** 
 * ConditionalBranchPStmt.java
 *
 * A two-way conditional branch statement. 
 *
 * Differs from an IfPStmt in that a ConditionalBranchPStmt does not contain
 * its "then" and "else" statements; it points to them via trueDest and
 * falseDest. Members trueGoto and falseGoto also point to trueDest and
 * falseDest, so they are redundant and may be removed later. Also, trueDest
 * and falseDest *are* next-linked into the main PProgram, unlike the thenStmt
 * and elseStmt of an IfPStmt, so they don't need to be refined when refining
 * a ConditionalBranchPStmt.
 * 
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class ConditionalBranchPStmt extends PStmt
{
  PCond cond;

  GotoPStmt trueGoto;
  GotoPStmt falseGoto;

  public ConditionalBranchPStmt
  (PFunctionDef parent, Expr sourceExpr, 
   PCond _cond, GotoPStmt _trueGoto, GotoPStmt _falseGoto)
  {
    super (parent, sourceExpr);
    cond = _cond;
    trueGoto = _trueGoto;
    falseGoto = _falseGoto;
  }

  public PCond getCond ()
  {
    return cond;
  }

  public void setCond (PCond _cond)
  {
    cond = _cond;
  }

  public GotoPStmt getTrueGoto ()
  {
    return trueGoto;
  }

  public GotoPStmt getFalseGoto ()
  {
    return falseGoto;
  }

  public PStmt getTrueDest ()
  {
    return trueGoto.getDest ();
  }

  public PStmt getFalseDest ()
  {
    return falseGoto.getDest ();
  }

  public void printMe (PrintWriter out)
  {
    out.print (getId () + "/" + getLabel () + ": ");
    out.println ("condExpr (" + cond + ")");
    out.println ("true goto " + trueGoto.getDest ().getId ()
	+ "\nfalse goto " + falseGoto.getDest ().getId ());
  }
}
