package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.util.IndentPrintWriter;


import edu.toronto.cs.yasm.abstractor.*;

import java.util.*;
import java.io.*;


/**
 * Describe class PrintingVisitor here.
 *
 *
 * Created: Mon Jun 20 14:51:36 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class PrintingVisitor extends BasePProgramVisitor 
{

  static int stmtCounter = 1;

  Map labelsForGoto;
  

  IndentPrintWriter out;

  /**
   * Creates a new <code>PrintingVisitor</code> instance.
   *
   */
  public PrintingVisitor (Writer _out) 
  {
    this (new IndentPrintWriter (_out));
    labelsForGoto = new HashMap();
    
  }

  public PrintingVisitor (IndentPrintWriter _out)
  {
    out = _out;
    labelsForGoto = new HashMap();
  }

  /**
   * Describe <code>visitPProgram</code> method here.
   *
   * @param p a <code>PProgram</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPProgram (PProgram p, Object o)
  {

    //set up labelsForGoto
    for (Iterator it = p.getStmtList ().iterator (); it.hasNext ();)
      {
	PStmt s = (PStmt) it.next ();

	if (s instanceof GotoPStmt)
	  {
	    GotoPStmt gotoS = (GotoPStmt) s;
	    labelsForGoto.put (gotoS.getTargetLabel (), new Boolean(true));
	  }
      }

    //visit the predicate-program
    out.println ("/* program starts */");
    out.println ("/* GLOBALS BEGIN */");
    if (!p.getGlobalDecls ().isEmpty ())
      {

	out.println ("{");
	out.indent ();
      
	for (Iterator it = p.getGlobalDecls ().values ().iterator ();
	     it.hasNext ();)
	  ((PDecl) it.next ()).accept (this, o);

	out.outdent ();
	out.println ("}");
      }
    
    
    out.println ("/* GLOBALS END */");
    /* print functions one after another, 
       although we may want to print main first or last */
    for (Iterator it = p.getFunctionDefs ().values ().iterator (); 
	 it.hasNext (); )
      {
	((PFunctionDef) it.next ()).accept (this, o);
	if (it.hasNext ()) out.println ();
      }

    out.flush ();
    return o;
  }
  

  /**
   * Describe <code>visitFunctionDef</code> method here.
   *
   * @param fdef a <code>PFunctionDef</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitFunctionDef (PFunctionDef fdef, Object o)
  {
    out.println (fdef.getFunctionName ());
    out.println ("{");
    out.indent ();
    out.println ("/* LOCALS */");

    if (!fdef.getLocalDecls ().isEmpty ())
      {

	out.println ("{");
	out.indent ();


	for (Iterator it = fdef.getLocalDecls ().values ().iterator (); 
	     it.hasNext ();)
	  ((PDecl) it.next ()).accept (this, o);

	out.outdent ();
	out.println ("}");
      }
    
    out.println ("/* FUNCTION BODY */");


    //visit each statment of the function
    //ignore the head (entry) and return selector statement?
    if ( fdef.getHead () != null)
      {	
	PStmt s;

	// -- starting from the entry point
	//    entryPoint = getHead ().getNext ();
	s = fdef.getEntryPoint (); 
	
	while (s != null)
	  {

	    // 	    	    if (s instanceof WhilePStmt)
	    // 	    	      out.println ("***** WhilePStmt:    ");    
	    // 	    	    else if (s instanceof IfPStmt)
	    // 	    	      out.println ("***** IfPStmt:    ");
	    // 	    	    else if (s instanceof BreakPStmt)
	    // 	    	      out.println ("***** BreakPStmt:    ");
	    // 	    	    else if (s instanceof GotoPStmt)
	    // 	    	      out.println ("***** GotoPStmt:    ");    
	    // 	    	    else if (s instanceof ReturnSelectorPStmt)
	    // 	    	      out.println ("***** ReturnSelectorPStmt:    ");    
	    // 	    	    else if (s instanceof NDGotoPStmt)
	    // 	    	      out.println ("***** NDGotoPStmt:    ");    
	    // 	    	    else if (s instanceof SkipPStmt)
	    // 	    	      out.println ("***** SkipPStmt:    ");   
	    // 	    	    else if (s instanceof FunctionCallPrologue)
	    // 	    	      out.println ("***** FunctionCallPrologue:    ");   
	    // 	    	    else if (s instanceof FunctionCallEpilogue)
	    // 	    	      out.println ("***** FunctionCallEpilogue:    ");   
	    // 	    	    else if (s instanceof ReturnPStmt)
	    // 	    	      out.println ("***** ReturnStmt:    ");   
	    // 	    	    else if (s instanceof ReturnSelectorPStmt)
	    // 	    	      out.println ("***** ReturnSelectorPStmt:    ");   
	    // 	    	    else if (s instanceof FunctionCallPStmt)
	    // 	    	      out.println ("***** FunctionCallPStmt:    ");    
	    // 	    	    else if (s instanceof PrllAsmtPStmt)
	    // 	    	      out.println ("***** PrllAsmtPStmt:    ");    

	    // out.println("***** stmtCounter = " + stmtCounter++);   

	    s.accept (this, o);
	    s = s.getNext ();
	    if (s != null) 
	      out.println ();
	  }
      }    

    // -- finally, visit ReturnSelector, the head statment of the defination
    fdef.getHead ().accept (this, o);

    out.outdent ();
    out.println ("}");
    return o;
  }

  public Object visitPStmt (PStmt s, Object o)
  {    

    checkLabelForGoto(s);   
    out.println (s.toString ());    

    return o;
  }
  

  /**
   * Print <code>IfPStmt</code>
   * <p>
   * first visit <code>cond</code>, 
   * then visit <code>thenStmt</code> and <code>elseStmt</code>
   * <p>
   * when dealing with  <code>thenStmt</code> and <code>elseStmt</code>, 
   * both are "manully" set Goto statment, we first get the corresponding 
   * statements using the labels, and then call accept to visit them
   */
  public Object visitIf (IfPStmt s, Object o)
  {
 
    checkLabelForGoto(s);

    out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");
    
    out.println ("if ( " + s.getCond (). getCond () + " )");
    out.println ("{");
    out.indent ();
    if (s.getThenGoto () != null)
      {
	PStmt ts;

	ts = s.getThenStmt ();

	//go through the statements corresponding  to (starting from)
	//the target of thenGoto
	while (ts != null)
	  {
	    ts.accept (this, o);
	    ts = ts.getNext();
	    if (ts != null) 
	      out.println ();
	  }
      }

    out.outdent ();
    out.println ("}");
    if (s.getElseGoto () != null)
      {
      
	out.println ("else");
	out.println ("{");
	out.indent ();

	PStmt ts;

	ts = s.getElseStmt ();

	//go through the statements corresponding  to (starting from)
	//the target of elseGoto	
	while (ts != null)
	  {
	    ts.accept (this, o);
	    ts = ts.getNext();
	    if (ts != null) 
	      out.println ();
	  }

	out.outdent ();
	out.println ("}");
      
      }

    
    return o;
    
  }


  /**
   * Print <code>WhilePStmt</code>
   */
  public Object visitWhile (WhilePStmt s, Object o)
  {

    checkLabelForGoto(s);

    out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");    
    
    out.println ("while ( " + s.getCond (). getCond () + " )");
    out.println ("{");
    out.indent ();
    if (s.getThenGoto () != null)
      {
	PStmt ts;

	// -- Print the "then" statement. 
	ts = s.getThenStmt ();

	//go through the statements corresponding  to (starting from)
	//the target of thenGoto
	while (ts != null)
	  {    
	    ts.accept (this, o);
	    ts = ts.getNext();
	    if (ts != null) 
	      out.println ();
	  }
      }

    out.outdent ();
    out.println ("}");

    
    return o;
    
  }


  /**
   * Print <code>GotoPStmt</code>
   */
  public Object visitGoto (GotoPStmt s, Object o)
  {

    checkLabelForGoto(s);

    out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");    

    out.println ("goto " + s.getTargetLabel () + ";");

    return o;
     
  }
  


  /**
   * Print <code>PrllAsmt</code>
   */
  public Object visitPrllAsmt (PrllAsmtPStmt s, Object o)
  {
    checkLabelForGoto(s);
    if (!s.getAsmts ().isEmpty ())
      {
	out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");    

	out.println ("{");
	out.indent ();

	for (Iterator it = s.getAsmts ().iterator (); it.hasNext (); )
	  {
	    AsmtPStmt asmt = (AsmtPStmt) it.next ();
	    if (!(asmt instanceof NDAsmtPStmt) && asmt.getVar () != null)
	      {    
		out.println (asmt.getVar () + " := ( " + asmt.getTrueCond () + 
			     " , " + asmt.getFalseCond () + " );");
	      }	
	    else if (asmt instanceof NDAsmtPStmt && asmt.getVar () != null)
	      // XXX Fix this
	      out.println (asmt.getVar () + "  := {true, false}");
	  }
    
	out.outdent ();
	out.println ("}");
      }
    
    return o;
     
  }
  

  /**
   * Print <code>FunctionCallPrologue</code>
   */
  public Object visitFunctionCallPrologue (FunctionCallPrologue s, Object o)
  {

    checkLabelForGoto(s);

    out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");    

    //print function name
    out.println ("[ " + s.getCall ().getFunctionName () + ".prologue ]");
    out.println ("{");
    out.indent ();

    //print the assignment corresponding to this function call
    for (Iterator it = s.getAsmts ().iterator (); it.hasNext (); )
      {
	AsmtPStmt asmt = (AsmtPStmt) it.next ();
	if (asmt.getVar () != null)
	  {    
	    out.println (asmt.getVar () + " := ( " + asmt.getTrueCond () + 
			 " , " + asmt.getFalseCond () + " );");
	  }
	
      }
    
    out.outdent ();
    out.println ("}");


    return o;
     
  }


  /**
   * Print <code>FunctionCallEpilogue</code>
   */
  public Object visitFunctionCallEpilogue (FunctionCallEpilogue s, Object o)
  {
    checkLabelForGoto(s);

    out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");    

    //print function name
    out.println ("[ " + s.getCall ().getFunctionName () + ".Epilogue ]");
    out.println ("{");
    out.indent ();

    //print the assignment corresponding to this function call
    for (Iterator it = s.getAsmts ().iterator (); it.hasNext (); )
      {
	AsmtPStmt asmt = (AsmtPStmt) it.next ();
	if (asmt.getVar () != null)
	  {    
	    out.println (asmt.getVar () + " := ( " + asmt.getTrueCond () + 
			 " , " + asmt.getFalseCond () + " );");
	  }
	
      }
    
    out.outdent ();
    out.println ("}");


    return o;
     
  }


  /**
   * Print <code>SkipPStmt</code>
   */
  public Object visitSkip (SkipPStmt s, Object o)
  {
   
    checkLabelForGoto(s);

    out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");    
    out.println ("skip;");

    return o;
     
  }

  /**
   * Print <code>BreakPStmt</code>
   */
  public Object visitBreak (BreakPStmt s, Object o)
  {


    checkLabelForGoto(s);

    out.println ("***** "+ s.getId () + "/" + s.getLabel () + " *****");    

    out.println ("goto " + s.getTargetLabel () + ";");

    return o;
        
  }


  /**
   * Print <code>PDecl</code>
   * <p>
   * print the datatype and variable name 
   * (no sure if this is necessary)
   */
  public Object visitDecl (PDecl d, Object o)
  {
    Expr sourceExpr;
    
    sourceExpr = d.getSourceExpr ();

    //sourceExpr.arg (0).arg (0) : datatype
    out.println (sourceExpr.arg (0).arg (0) + " : " + 
		 CILDeclarationOp.getDeclName (sourceExpr) + ";");

    return o;

  }


  /**
   * Print the label of <code>PStmt</code> if
   * it is used by some goto statement
   */
  private void checkLabelForGoto (PStmt s)
  {
    if (labelsForGoto.containsKey (s.getLabel ()))
      {
	out.print(s.getLabel () + ":");
	out.println();
      }
    
  }



  public static void main (String[] args) throws Exception
  {
    PProgram pProgram = 
      PProgram.parse 
      (new NullExprAbstractor (new ExprFactoryImpl ()), args [0]);
 
    PrintWriter myout = new PrintWriter(System.out);
    pProgram.print(myout);
    myout.flush ();

    pProgram.accept 
      (new PrintingVisitor (new IndentPrintWriter (System.out)), null);
    
  }
}
