package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.expr.*;

import java.util.*;
import java.io.*;

public class ReturnSelectorPStmt extends PStmt
{
  List<PStmt> dests;
  PredicateTable.IntVariable returnIndex;

  public ReturnSelectorPStmt (PFunctionDef parent)
  {
    super (parent, (Expr) null);
    dests = new ArrayList (parent.getNumCallSites ());
  }

  // constructor which does not build dests or returnIndex
  public ReturnSelectorPStmt (PFunctionDef parent, boolean dummy)
  {
    super (parent, (Expr) null);
    dests = new LinkedList ();
  }

  public PredicateTable.IntVariable getReturnIndex ()
  {
    if (returnIndex != null)
      return returnIndex;

    returnIndex = getPProgram ().getPredTable ().newIntVar
      (parent.getReturnIndexVar (), dests.size ());

    return returnIndex;
  }

  /** 
   * Constructs and returns a list of return site labels.
   * 
   * @return 
   * @deprecated Use {@code getDests ()}.
   */
  public List<String> getTargetLabels ()
  {
    List<String> labels = new ArrayList<String> (parent.getNumCallSites ());

    for (int i = 0; i < parent.getNumCallSites (); i++)
      labels.add (parent.returnSiteLabel (i));

    return labels;  
  }

  public List<PStmt> getDests ()
  {
    return dests;
  }

  /** 
   * For a return selector with less than two call sites, returns a 
   * GotoPStmt with the same id which jumps to END or the unique dest when
   * there are zero or one call sites, respectively; the goto's next and dest
   * are not set.
   */
  public GotoPStmt asGoto ()
  {
    assert dests.size () < 2 : "Can't treat this as GotoPStmt with >= 2 dests";
    GotoPStmt g = new GotoPStmt ();
    g.setId (getId ());
    if (dests.size () == 0)
      g.setDest (getParent ().getParent ().getStmt ("END"));
    else
      g.setDest (dests.get (0));
    assert g.getDest () != null;
    return g;
  }

  public void addDest (PStmt dest)
  {
    //assert (returnIndex == null);
    // do not add dests after computing/getting returnIndex
    dests.add (dest);
  }

  public Object accept (PProgramVisitor v, Object o)
  {
    return v.visitReturnSelector (this, o);
  }
  

  public void printMe (PrintWriter out)
  {
    out.println (toString ());
  }

  public String toString ()
  {
    StringBuffer sb = new StringBuffer ();
    sb.append (getId () + "/" + getLabel () + ": ");
    sb.append ("Return dest IDs [");
    for (PStmt p : getDests ())
      sb.append (p.getId () + " ");
    sb.append ("]");
    return sb.toString ();
  }
}
