package edu.toronto.cs.yasm.reachability;

import edu.toronto.cs.mvset.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.yasm.pcfa.*;

/**
 * A path proof rule. Reduces PathProofGoal into a BoundedPathProofGoal
 * 
 *
 *
 * Created: Fri Jun 10 16:27:02 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class PathRule extends ReachabilityProofRule 
{

  MvSet stateCube;
  
  /**
   * Creates a new <code>PathRule</code> instance.
   *
   */
  public PathRule (BelnapAlgebra _algebra, 
		    MvSet _stateCube) 
  {
    super (_algebra);
    stateCube = _stateCube;
  }

  public ProofGoal[] apply  (ProofGoal _goal)
  {

    if (! (_goal instanceof PathProofGoal)) return null;
    
    PathProofGoal goal = (PathProofGoal) _goal;

    // -- current set of nodes we are working with, starting
    // -- with what is in the proof goal
    NodeSet nodes = goal.getNodes ();
    // -- current value we computed, starting with bot
    AlgebraValue val  = bot;
    do 
      {
	nodes = nodes.getPrev ();
	Node valNode = nodes.get (goal.getState ().getId ());

	if (valNode == null)
	  val = bot;
	else
	  val = valNode.getValue ().and (goal.getState ().getValue ()).
	    existAbstract (stateCube).getValue ();
      }
    while (val.equals (goal.getValue ()) && nodes != null);

    if (nodes == null) return null;
    
    
    return oneGoal (new BoundedPathProofGoal (goal, 
					      nodes.getNext (), 
					      goal.getValue (),
					      goal.getState ()));
  }

  public String toString ()
  {
    return "PathRule";
  }
  
  

}
