package edu.toronto.cs.yasm.reachability;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.yasm.pcfa.*;

/**
 * Describe class SimpleMinPathRule here.
 *
 *
 * Created: Mon Jun 13 10:59:20 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class SimpleMinPathRule extends ReachabilityProofRule {

  /**
   * Creates a new <code>SimpleMinPathRule</code> instance.
   *
   */
  public SimpleMinPathRule (BelnapAlgebra _algebra) 
  {
    super (_algebra);
  }

  public ProofGoal[] apply (ProofGoal _goal)
  {
//     if (2 == 2)
//       return null;
    
    if (! (_goal instanceof BoundedPathProofGoal)) return null;
    
    BoundedPathProofGoal goal = (BoundedPathProofGoal) _goal;
    
    NodeSet curr = goal.getNodes ();
    NodeSet prev = curr.getPrev ();
    
    Node.NodeId goalId = goal.getState ().getId ();

    // -- skip back while value at goalId node is unchanged
    while (prev != null && prev.get (goalId) != null)
      {
	if (curr.get (goalId).getValue ().equals
	    (prev.get (goalId).getValue ()))
	  {
	    curr = prev;
	    prev = curr.getPrev ();
	  }
	else
	  break;
      }
    
    if (curr == goal.getNodes ())
      // -- made no progress
      return null;
    
    System.err.println ("Simple Min Path reduced " + goal.getBound () + 
			" to " + curr.getCount ());
      
    return oneGoal (new BoundedPathProofGoal (goal, 
					      curr,
					      goal.getValue (),
					      curr.get 
					      (goal.getState ().getId ())));
					      
  }
  

}
