package edu.toronto.cs.yasm.reachability;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.yasm.pcfa.*;

import java.util.Iterator;


/**
 * Reduces BoundedPathGoal by one step
 *
 *
 * Created: Mon Jun 13 11:15:06 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class StepRule extends ReachabilityProofRule 
{

  EdgeSet edges;
  MvSetFactory mvFactory;
  MvSet stateCube;
  
  
  /**
   * Creates a new <code>StepRule</code> instance.
   *
   */
  public StepRule (BelnapAlgebra _algebra, EdgeSet _edges, MvSet _stateCube) 
  {
    super (_algebra);
    edges = _edges;
    stateCube = _stateCube;
    mvFactory = stateCube.getFactory ();
  }

  public ProofGoal[] apply (ProofGoal _goal)
  {
    if (! (_goal instanceof BoundedPathProofGoal))
      return null;
    
    BoundedPathProofGoal goal = (BoundedPathProofGoal) _goal;
    
    System.err.println ("Applying step rule with goal.getBound: " + 
			goal.getBound ());
    
    if (goal.getBound () == 0) return null;
    

    NodeSet phi = goal.getNodes ().getPrev ();


    // -- get the "final" (the last computation for EX) value of Phi
    // -- use it to determine the value of next state
    NodeSet finalPhi;
    
    finalPhi = phi;

    while(finalPhi.getNext () != null)
      {
	finalPhi = finalPhi.getNext ();
      }
    
    
    NodeSet srcSet = new NodeSet ();
    srcSet.add (goal.getState ());
    System.err.println ("Starting to look at edges from " + goal.getState ().getId ());
    for (Iterator it = 
	   edges.edgesFrom (goal.getState ().getId ()).iterator ();
	 it.hasNext ();)
      {
	Edge edge = (Edge) it.next ();
	System.err.println ("Doing edge: " + edge + " of type " + 
			    edge.getClass ());
	NodeSet out = new NodeSet ();
	edge.fwdImage (srcSet, out);
	for (Iterator jt = out.nodes ().iterator (); jt.hasNext ();)
	  {
	    Node outNode = (Node) jt.next ();
	    Node phiNode = phi.get (outNode.getId ());

	    Node finalPhiNode = finalPhi.get (outNode.getId ());

	    
	    // -- skip an empty node
	    if (phiNode == null || finalPhiNode == null) 
	      {
		System.err.println ("\t skipping empty phi node: " + 
				    outNode.getId ());
		continue;
	      }
	    
	    
	    MvSet transAndPhi = outNode.getValue ().and (phiNode.getValue ());
	    
	    // -- skip false conjuncts
	    if (transAndPhi.isConstant () && 
		transAndPhi.getValue ().equals (bot))
	      {
		System.err.println ("Got an empty conjunct on: " + 
				    outNode.getId ());
		continue;
	      }
	    

	    // -- ok, finally have something
	    Node dstState = 
	      new Node (outNode.getId (),
			(MvSet) transAndPhi.
			mintermIterator (mvFactory.top (), 
					 goal.getValue ()).next ());
	    
// 	    AlgebraValue phiVal = 
// 	      phiNode.getValue ().and (dstState.getValue ()).
// 	      existAbstract (stateCube).getValue ();


	    AlgebraValue phiVal = 
	      finalPhiNode.getValue ().cofactor (dstState.getValue ()).
	      existAbstract (stateCube).getValue ();


// 	    System.err.println (" ****** value at this state : " + phiVal);
	    

	    // XXX this just does not look right!
	    AlgebraValue transVal = outNode.getValue ().cofactor 
	      (dstState.getValue ()).existAbstract (stateCube).getValue ();
	    
	    
	    System.err.println ("1Done looking at edges from " + goal.getState ().getId ());

// 	    System.out.println ("from state: " + goal.getState ());
// 	    System.out.println ("to state: " + dstState);
	    return 
	      twoGoal (new BoundedPathProofGoal (goal, phi, phiVal, dstState),
		       new EdgeProofGoal (goal, goal.getState (), 
					  dstState, edge, transVal));
	  }
	
      }
    System.err.println ("2Done looking at edges from " + goal.getState ().getId ());

    
    
    return null;
  }
  

}
