package edu.toronto.cs.yasm.refiner;

import java.util.*;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.*;
import edu.toronto.cs.cparser.*;
import edu.toronto.cs.util.*;

import static edu.toronto.cs.cparser.CILTransformerUtil.*;
import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;
import static edu.toronto.cs.expr.NumericOp.*;


import edu.toronto.cs.mvset.*;
import edu.toronto.cs.algebra.TwoValAlgebra;
import edu.cmu.sei.pacc.absdom.*;

public class AbstractorTest
{
  ExprFactory eFac = new ExprFactoryImpl ();
  
  
  public void testRun ()
  {
    Expr i = eFac.var ("i");
    Expr j = eFac.var ("j");
    Expr _j = eFac.var ("_j");
    Expr src = eFac.var ("src");
    Expr dest = eFac.var ("dest");
    
    Expr intT = eFac.op (INT);
    
    SymbolTable st = new SymbolTable ();
    
    st.put (i, intT);
    st.put (j, intT);
    st.put (_j, intT);
    st.put (src, arrayT (intT, intT));
    st.put (dest, arrayT (intT, intT));
    
    List<Expr> preds = new LinkedList<Expr> ();
   
    try
      {
	preds.add (getExpr (eFac, "i <= 50"));
	preds.add (getExpr (eFac, "i < 50"));
	preds.add (getExpr (eFac, "src[i] == 0"));
	preds.add (getExpr (eFac, "src [50] == 0"));
	preds.add (getExpr (eFac, "j <= 50"));
	preds.add (getExpr (eFac, "j < 50"));
	preds.add (getExpr (eFac, "dest[j] == 0"));
	preds.add (getExpr (eFac, "dest[50] == 0"));
	preds.add (getExpr (eFac, "_j <= i"));
	preds.add (getExpr (eFac, "_j <= 50"));
	preds.add (getExpr (eFac, "0 <= i"));
	preds.add (getExpr (eFac, "j < 101"));
      }
    catch (Exception ex)
      {
	throw new RuntimeException (ex);
      }
    
    
    System.out.println ("Our predicates are: ");
    int count = 0;
    for (Expr e : preds)
      System.out.println ((count++) + ": " + e);


    ITheoremProver tp =
      new CTheoremProver (eFac, new CLCVCLTheoremProver (eFac, true), st);
 

    TPCBJPredicateAbstractor cbj = 
      new GNGTPCBJPredicateAbstractor (preds, eFac.falseExpr (), tp);

    System.out.println ("Running GNG cbj");
    cbj.recSearch (1);
    
    Expr tExpr = cbj.getTExpr ();
    System.out.println ("tExpr: " + tExpr);
    //System.out.println ("fExpr: " + cbj.getFExpr ());

    System.out.println ("disj in tExpr: " + countDisj (tExpr));


    MvSetAbstractDomain aDom = 
      new MvSetAbstractDomain 
      (MDDMvSetFactory.newMvSetFactory (new TwoValAlgebra (), 12), eFac);


    // -- uncomment to force the same order in MvSet as in CBJ
    //for (Expr p : preds) aDom.alpha (p);

    MvSet mvSet = aDom.alpha (tExpr);
    Expr simp = aDom.gamma (mvSet);

    
    System.out.println ("disj in simp: " + countDisj (simp));
    System.out.println ("nodes in mvSet: " + mvSet.size ());
    

    //cbj.dump ();

    TPCBJPredicateAbstractor cbj2 = 
      new TPCBJPredicateAbstractor (preds, eFac.falseExpr (), tp);

    System.out.println ("Running TPCBJ");
    cbj2.recSearch (1);
    
    System.out.println ("Checking semantic equivalence of new and old");
    System.out.println ("cbj -> cbj2: " + 
			tp.implies (cbj.getTExpr (), cbj2.getTExpr ()));
    System.out.println ("cbj2 -> cbj: " + 
			tp.implies (cbj2.getTExpr (), cbj.getTExpr ()));

    
    MvSet mvSet2 = aDom.alpha (cbj2.getTExpr ());

    System.out.println ("Checking whether new and old result " + 
			"in the same MvSet representation (they should)");
    
    System.out.println ("mvset == mvset2: " + mvSet.equals (mvSet2));
    
  }
  

  public static void main (String[] args)
  {
    new AbstractorTest ().testRun ();
  }


  static int countDisj (Expr e)
  {
    if (e.op () != BoolOp.OR) return 1;

    int sum = 0;
    
    for (Expr arg : e.args ())
      sum += countDisj (arg);
    return sum;
  }
  
}
