package edu.toronto.cs.yasm.refiner;

import edu.toronto.cs.tp.*;
import edu.toronto.cs.expr.*;

import java.util.IdentityHashMap;
import java.util.Map;
import java.util.List;
import java.util.LinkedList;
import java.util.BitSet;

import edu.toronto.cs.algebra.TwoValAlgebra;
import edu.toronto.cs.mvset.*;
import edu.cmu.sei.pacc.absdom.MvSetAbstractDomain;



/**
 * Global No-Good TP CBJ Predicate Abstractor.
 * (sorry for the name)
 */
public class GNGTPCBJPredicateAbstractor extends TPCBJPredicateAbstractor
{

  /**  mv-set encoder */
  MvSetAbstractDomain enc;

  MvSetFactory mFac;
  

  /** current global no-good that is being constructed */
  BitSet[] currGlobalNG;
  
  /** all seen assignments */
  MvSet[] currSeen;
  
  ExprFactory eFac;

  boolean isTargetFalse;
  
  
  public GNGTPCBJPredicateAbstractor (List _srcPredicates, 
				      Expr _targetPredicate, 
				      ITheoremProver _tp)
  {
    super (_srcPredicates, _targetPredicate, _tp);

    eFac = _targetPredicate.getFactory ();
    
    enc = new MvSetAbstractDomain 
      (MDDMvSetFactory.newMvSetFactory (new TwoValAlgebra (), 10), 
       eFac); 

    mFac = enc.getMvSetFactory ();
    
    currGlobalNG = new BitSet [variables.length + 1];

    // -- at level 0, put empty no-good
    currGlobalNG [0] = new BitSet ();
    currGlobalNG [0].set (0);

    currSeen = new MvSet [variables.length + 1];
    // -- initialy we saw nothing, empty set, false
    currSeen [0] = enc.alpha (eFac.falseExpr ());

  
    isTargetFalse = (eFac.falseExpr () == target);
  }
  

  
  void assign (CSPVariable<Expr> var, int val, int level)
  {
    super.assign (var, val, level);
    updateGNG (level);

    // -- for debugging
//     if (mFac.top ().equals (currSeen [level]))
//       {
// 	System.out.println ("Saw true at level: " + level);
// 	System.out.println ("One level up is: " + 
// 			    enc.gamma (currSeen [level - 1]));
	
//       }

  }

  void updateGNG (int level)
  {
    // -- restrict "seen" by the current assignment
    currSeen [level] = currSeen [level - 1].cofactor 
      (mvSetForVar (currentAssignment [level]));
    
    // -- copy global no good
    currGlobalNG [level] = (BitSet) currGlobalNG [level - 1].clone ();
    

    

    // -- if current assignment reduced anything, assume it was important
    // -- and set this into the no-good
    if (! currSeen [level].equals (currSeen [level - 1]))
      currGlobalNG [level].set (level);
  }
  
  
  public BitSet checkConsistency (CSPVariable<Expr> curVar)
  {
    // -- see if we hit a global no-good first
    if (mFac.top ().equals (currSeen [curVar.getCurLevel ()]))
      {
// 	System.out.println ("Early termination at " + curVar + " val " + 
// 			    curVar.getCurVal () + " at depth " + 
// 			    curVar.getCurLevel ());
// 	System.out.println ("No good is: " + 
// 			    currGlobalNG [curVar.getCurLevel ()]);
	
	return currGlobalNG [curVar.getCurLevel ()];
      }
    
    
    // -- drop into check-consistency 
    return super.checkConsistency (curVar);
  }
  

  protected void newSolutionEvent (Boolean result, Expr solution)
  {

    /**
     * A hack to get arround that we may be in a situation where
     * target /\ !target == false but target =&gt; !target
     */
    if (isTargetFalse &&  result != Boolean.TRUE) return;
    
    
    // -- convert solution into an mvset
    MvSet newSeen = enc.alpha (solution);
    
    // -- add to the list of what we saw so far
    currSeen [0] = currSeen [0].or (newSeen);

//     System.out.println ("Adding " + solution + " with value " + result);
//     System.out.println ("Got: " + enc.gamma (currSeen [0]));

    // -- update currSeen up to the bt-level
    for (int i = 1; i <= assignedVariables; i++)
      updateGNG (i);    
  }
  

  /**
   * MvSet for a variable assignment
   */
  protected MvSet mvSetForVar (CSPVariable<Expr> var)
  {
    return mvSetForVar (var, var.getCurVal ());
  }
  
  protected MvSet mvSetForVar (CSPVariable<Expr> var, int val)
  {
    MvSet res = enc.alpha (var.getExpr ());
    if (val == 0) res = res.not ();
    return res;
  }


  public void dump ()
  {
    //System.out.println ("currSeen at the end is: " );
    //System.out.println (enc.gamma (currSeen [0]));
  }
  
  
}
