package edu.toronto.cs.yasm.refiner;

import java.util.*;

import junit.framework.*;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.*;
import edu.toronto.cs.cparser.*;
import edu.toronto.cs.util.*;

import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;
import static edu.toronto.cs.expr.NumericOp.*;

public class TPCBJPredicateAbstractorTestCase extends TestCase
{
  private ExprFactory fac;
  private Expr intT;

  protected void setUp ()
  {
    fac = new ExprFactoryImpl ();
    intT = fac.op (INT);
  }

  public void testSuperSimpleBug () throws Exception
  {
    Expr dest = fac.var ("dest");
    Expr buf_ext = fac.var ("buf_ext");
    Expr buf = fac.var ("buf");
    Expr src = fac.var ("src");
    Expr input = fac.var ("input");

    Expr charT = fac.op (CHAR);

    SymbolTable st = new SymbolTable ();
    st.put (dest, ptrT (charT)); // char *dest
    st.put (buf_ext, ptrT (charT)); // char *buf_ext
    st.put (buf, arrayT (intT, charT)); // char buf[]
    st.put (input, arrayT (intT, charT)); // char input[]
    st.put (src, ptrT (charT)); // char *src

//    System.out.println ("Symbol table: " + st);

    List<Expr> preds = new LinkedList<Expr> ();

    Expr p1 = CILTransformerUtil.getExpr (fac, "dest > buf_ext");
    Expr p2 = CILTransformerUtil.getExpr (fac, "&buf[0] > buf_ext");
    Expr p3 = CILTransformerUtil.getExpr (fac, "*src == 0");
    Expr p4 = CILTransformerUtil.getExpr (fac, "*(src + 1) == 0");
    Expr p5 = CILTransformerUtil.getExpr (fac, "dest + 1 > buf_ext");

    preds.add (p1);
    preds.add (p2);
    preds.add (p3);
    preds.add (p4);
    preds.add (p5);

//    System.out.println ("preds: " + preds);

    // target is wp (input [0] := 1, *src == 0)

    Expr target =
      ChoiceOp.choice (
          CILTransformerUtil.getExpr (
            fac, "!(&input[0] == src) && (*src == 0)"),
          CILTransformerUtil.getExpr (
            fac, "&input[0] == src || (!(&input[0] == src) && !(*src == 0))"));

//    System.out.println ("target: " + target);

    // CLCVCL logging enabled
    ITheoremProver tp =
      new CTheoremProver (fac, new CLCVCLTheoremProver (fac, true), st);

    TPCBJPredicateAbstractor cbj =
      new TPCBJPredicateAbstractor (preds, target, tp);

    cbj.recSearch (1);

    SystemOutWrapper.writer.println ("true approx: " + cbj.getTExpr ());
    SystemOutWrapper.writer.println ("false approx: " + cbj.getFExpr ());
  }

  /** 
   * A test to see if CBJ checks whether a context is consistent before doing a
   * query. 
   *
   * Hrm, looks like CBJ generates inconsistent contexts.
   * The output from this says that y > 0 is valid under the inconsistent
   * context x > 0 AND NOT(x + 1 > 0).
   */
  public void testInconsistency () throws Exception
  {
    Expr x = fac.var ("x");
    Expr y = fac.var ("y");

    SymbolTable st = new SymbolTable ();
    st.put (x, intT);
    st.put (y, intT);

    Expr p1 = CILTransformerUtil.getExpr (fac, "x > 0");
    Expr p2 = CILTransformerUtil.getExpr (fac, "x + 1 > 0");

    List<Expr> preds = new LinkedList<Expr> ();
    preds.add (p1);
    preds.add (p2);

    Expr target = CILTransformerUtil.getExpr (fac, "y > 0");

    ITheoremProver tp =
      new CTheoremProver (fac, new CLCVCLTheoremProver (fac, true), st);

    TPCBJPredicateAbstractor cbj =
      new TPCBJPredicateAbstractor (preds, target, tp);

    cbj.recSearch (1);

    SystemOutWrapper.writer.println ("true approx: " + cbj.getTExpr ());
    SystemOutWrapper.writer.println ("false approx: " + cbj.getFExpr ());
  }
}

