package edu.toronto.cs.yasm.wp;

import edu.toronto.cs.expr.*;
import java.util.List;

/**
 * An interface representing a weakest precondition computer.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public interface WPComputer
{

  /**
   * computes the weakest pre-condition of {@code expr}
   *
   * @param expr a post-condition to establish
   * @return an a weakest pre-condition guaranteeing the
   * post-condition {@code expr}
   */
  Expr computeWP (Expr expr);
  
  /**
   * Pointwise extension of 
   * {@link edu.toronto.cs.yasm.wp.WPComputer#computeWP(Expr)} to arrays of 
   * expressions. If {@code A} is an array of {@code Expr}, then 
   * {@code B = computeWP(A)} is such that {@code B[i] == computeWP(A[i])}
   *
   * @param expr an array of post-conditions
   * @return an array of weakest pre-conditions
   */
  Expr[] computeWP (Expr[] expr);
  
  /**
   * Pointwise extension of 
   * {@link edu.toronto.cs.yasm.wp.WPComputer#computeWP(Expr)} to lists of
   * expressions. If {@code A} is a list {@code Expr}, then 
   * {@code B = computeWP(A)} is such that 
   * {@code B.get(i) == computeWP(A.get(i))}
   *
   * @param expr a <code>List</code> of post-conditions
   * @return a <code>List</code> of weakest pre-conditions
   */
  List computeWP (List expr);

}
